unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, odbcsql, Forms, Controls, Graphics, Dialogs, StdCtrls, LazLogger, FileUtil;

Const
  UserName:    PSQLCHAR = nil;
  Password:    PSQLCHAR = nil;
  FormCaption: pChar = 'Form1';
  ODBC_MAX_CHAR_SIZE = 8000;   // I had problems with fields greater then 8000 characters!

type
  TTableField = Array[0..ODBC_MAX_CHAR_SIZE+1] of SQLCHAR; // +1 for null terminated string, no unicode/wide character support in this stage

  { TSQLExecThread }
  TSQLExecThread = class(TThread)
  private
    fStartTime:  TDateTime;
  protected
    procedure Execute; override;
  public
    Constructor Create(CreateSuspended: boolean);
  end;

  { TForm1 }
  TForm1 = class(TForm)
    Button1: TButton;
    ComboBox1: TComboBox;
    ComboBox2: TComboBox;
    procedure Button1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure OnThreadDone(Sender: TObject);
  private

  public
    SQLExecThread: TSQLExecThread;
  end;

var
  Form1: TForm1;

  DBDSN:      PSQLCHAR;
  EnvHandle:  SQLHandle;
  DBHandle:   SQLHandle;
  StmtHandle: SQLHSTMT;
  Column:     Array of TTableField; // Matches length of field+1
  ColCnt:     SQLSMALLINT;

  Query:   pchar;
  RowCnt:  Int64;
  StringLength2Ptr: SQLSMALLINT;


implementation

{$R *.lfm}

function ODBCSuccess (Res : Integer): Boolean;
begin
  ODBCSuccess := (res = SQL_SUCCESS) or (res = SQL_SUCCESS_WITH_INFO);
end;

procedure FreeHandles;
begin
  if assigned(StmtHandle) then
    SQLFreeHandle(SQL_HANDLE_STMT, StmtHandle);
  if assigned(dbhandle) then
    SQLFreeHandle(SQL_HANDLE_DBC, DBHandle);
  if assigned(EnvHandle) then
    SQLFreeHandle(SQL_HANDLE_ENV, EnvHandle);
  if assigned(Column) then
    setLength(Column, 0);
end;

procedure DoError (Msg: String; ErrCode: Integer; DoExit: Boolean);
begin
  FreeHandles;
  Writeln(Msg,' Code : ', ErrCode);
  if DoExit then
  begin
    FreeHandles;
    Halt(1);
  end;
end;

procedure StartSession;
var
  Res: Integer;
begin
  EnvHandle  := nil;
  DBHandle   := nil;
  StmtHandle := nil;
  Res := SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, EnvHandle);
  if Res <> SQL_SUCCESS then
    DoError('Could allocate ODBC handle', Res, true);
  Res := SQLSetEnvAttr(EnvHandle,SQL_ATTR_ODBC_VERSION, SQLPOINTER(SQL_OV_ODBC3), 0);
  If Not ODBCSuccess(res) then
    DoError('Could not set environment', Res, true);
  Res := SQLAllocHandle(SQL_HANDLE_DBC, envHandle, DBHandle);
  If res<>SQL_SUCCESS then
    DoError('Could not create database handle', res, true);
  Res := SQLConnect(DBHandle,
                     PSQLCHAR(DBDSN),    SQL_NTS,
                     PSQLChar(UserName), SQL_NTS,
                     PSQLCHAR(Password), SQL_NTS);
  {Res := SQLDriverConnect(DBHandle,
                          SQLHWND(GetDesktopWindow()),
                          DBDSN,
                          SQL_NTS,
                          nil,
                          0,
                          StringLength2Ptr,
                          SQL_DRIVER_COMPLETE);}
  if not OdbcSuccess(res) then
    DoError('Could not connect to datasource.', Res, true);
end;

procedure ExecuteStatement;
var
  Res,ErrCode, i : LongInt;
begin
  Res := SQLAllocHandle(SQL_HANDLE_STMT, DBHandle, StmtHandle);
  if not ODBCSuccess(res) then
    DoError('Could not allocate statement handle.', Res, true);

  // Now actually execute SQL Statement
  Res := SQLExecDirect(StmtHandle, Query, SQL_NTS);
  if not ODBCSuccess(res) then
    DoError('Execute of statement failed.', Res, false); //when statement is canceled, do not exit application

  // When Statement was executet we know the number of fields which where selected
  Res := SQLNumResultCols(StmtHandle, ColCnt);
  if not ODBCSuccess(res) then
    DoError('Execute of SQLNumResultCols failed.', Res, false)
  else
    setLength(Column, ColCnt);

  { Bind result buffers.
    Note that for many queries, the result is not known on beforehand,
    And must be queried with SQLPrepare, SQLNumResulCols and SQLDescribeCol
    before the statement is executed.}
  for i := 0 to ColCnt do
    SQLBindCol(StmtHandle, i+1, SQL_CHAR, SQLPointer(@Column[i]), ODBC_MAX_CHAR_SIZE, @ErrCode);
end;

function GetThreadIsRunning(aThread: TThread): Boolean;
begin
  if Assigned(aThread) then // Check if thread was already created
    Result := not aThread.Finished // Thread is running or is waiting for execution of function OnThreadDone
  else
    Result := False; // Thread wurde noch nicht gestartet oder ist bereits fertig
end;

{ TSQLExecThread }
constructor TSQLExecThread.Create(CreateSuspended: boolean);
begin
  inherited Create(CreateSuspended);
  FreeOnTerminate := True;
end;

procedure TSQLExecThread.Execute;
var
  Res:      Longint;
  Count, i: Int64;

begin
  StartSession;
  ExecuteStatement;

  //ShowResult;
  Res   := SQLFetch(StmtHandle);

  Count := 0;
  while (Res = SQL_SUCCESS) and (not Terminated) do
  begin
    inc(Count);
    if (Count mod 1) = 0 then
    begin
      write('Record: ', Count, ' : ');
      for i := 0 to ColCnt-2 do
        write(PSQLCHAR(@Column[i][0]) + ' ');
      writeln(PSQLCHAR(@Column[ColCnt-1][0]));
    end;
    Res := SQLFetch(StmtHandle);
  end;
  RowCnt := Count;
  //writeln('Cleanup - Not terminated');
  FreeHandles;
end;

{ TForm1 }
procedure TForm1.Button1Click(Sender: TObject);
var
  strList: TStringList;

begin
  DBDSN := PSQLCHAR(ComboBox2.Text);

  strList := TStringList.Create;
  strList.LoadFromFile(ComboBox1.Text);

  Query := pChar(strList.Text);
  if not GetThreadIsRunning(SQLExecThread) then
  begin
    SQLExecThread := TSQLExecThread.Create(True);
    SQLExecThread.OnTerminate := @OnThreadDone;
    SQLExecThread.fStartTime := Now;
    SQLExecThread.Start;
    Button1.Caption := 'Stop';
    Form1.Caption := FormCaption;
  end
  else
  begin
    Button1.Enabled := false;
    SQLCancel(StmtHandle);
    SQLExecThread.Terminate;
    Form1.Caption := FormCaption;
  end;
  Application.ProcessMessages;
  strList.Free;
end;

procedure TForm1.FormCreate(Sender: TObject);
var
  SQLFiles: TStringList;
begin
  SQLFiles := TStringList.Create;
  try
    FindAllFiles(SQLFiles, '.', '*.sql', false); //find e.g. all SQL sourcefiles
    ComboBox1.Items.Assign(SQLFiles);
    ComboBox1.ItemIndex := 0;
  finally
    SQLFiles.Free;
  end;
end;

procedure TForm1.OnThreadDone(Sender: TObject);
begin
  if (SQLExecThread.terminated) then
  begin
    Form1.Caption := 'Thread terminated. Runtime: ' + FormatDateTime('HH:NN:SS', Now - SQLExecThread.fStartTime) + '. Processed rows: ' + IntToStr(RowCnt)
  end
  else
  begin
    Form1.Caption := 'Thread completed. Runtime: ' + FormatDateTime('HH:NN:SS', Now - SQLExecThread.fStartTime) + '. Processed rows: ' + IntToStr(RowCnt);
  end;
  Button1.Enabled := true;
  Button1.Caption := 'Start';
  Application.ProcessMessages;
end;
end.

